@@ -1,5 +1,7 @@
 #!/usr/bin/env bash
 
+set -e
+
 # Set a fixed umask as this leaks into docker containers
 umask 0022
 
@@ -33,7 +35,7 @@ function verify_signature() {
 
 function verify_hash() {
     local file=$1 expected_hash=$2
-    actual_hash=$(sha256sum $file | awk '{print $1}')
+    actual_hash=$(sha256sum "$file" | awk '{print $1}')
     if [ "$actual_hash" == "$expected_hash" ]; then
         return 0
     else
@@ -45,8 +47,8 @@ function verify_hash() {
 
 function download_if_not_exist() {
     local file_name=$1 url=$2
-    if [ ! -e $file_name ] ; then
-        wget -O $file_name "$url"
+    if [ ! -e "$file_name" ] ; then
+        wget -O "$file_name" "$url"
     fi
 }
 
@@ -72,6 +74,24 @@ clone_or_update_repo() {
     fi
 }
 
+apply_patch() {
+    local patch=$1
+    local path=$2
+
+    if [ -z "$patch" ] || [ -z "$path" ]; then
+        fail "apply_patch: invalid arguments: patch='$patch', path='$path'"
+    fi
+
+    if [ -d "$path" ]; then
+        info "Patching: $patch"
+        cd "$path"
+        patch -p1 <"$patch"
+        cd -
+    else
+        fail "apply_patch: path='$path' not found"
+    fi
+}
+
 # https://github.com/travis-ci/travis-build/blob/master/lib/travis/build/templates/header.sh
 function retry() {
     local result=0
