@@ -29,7 +29,7 @@ import hashlib
 import re
 import copy
 from typing import Tuple, TYPE_CHECKING, Union, Sequence, Optional, Dict, List, NamedTuple, Any, Type
-from functools import lru_cache, wraps
+from functools import wraps
 from abc import ABC, abstractmethod
 
 import electrum_ecc as ecc
@@ -52,6 +52,7 @@ from .util import (InvalidPassword, WalletFileException,
 from .mnemonic import Mnemonic, Wordlist, calc_seed_type, is_seed
 from .plugin import run_hook
 from .logging import Logger
+from .lrucache import LRUCache
 
 if TYPE_CHECKING:
     from .gui.common_qt.util import TaskThread
@@ -398,6 +399,9 @@ class Deterministic_KeyStore(Software_KeyStore):
 
 class MasterPublicKeyMixin(ABC):
 
+    def __init__(self):
+        self._pubkey_cache = LRUCache(maxsize=10**4)  # type: LRUCache[Sequence[int], bytes]  # path->pubkey
+
     @abstractmethod
     def get_master_public_key(self) -> str:
         pass
@@ -434,8 +438,14 @@ class MasterPublicKeyMixin(ABC):
     def get_key_origin_info(self) -> Optional[KeyOriginInfo]:
         return None
 
-    @abstractmethod
     def derive_pubkey(self, for_change: int, n: int) -> bytes:
+        key = (for_change, n)
+        if key not in self._pubkey_cache:
+            self._pubkey_cache[key] = self._derive_pubkey(*key)
+        return self._pubkey_cache[key]
+
+    @abstractmethod
+    def _derive_pubkey(self, for_change: int, n: int) -> bytes:
         """Returns pubkey at given path.
         May raise CannotDerivePubkey.
         """
@@ -501,6 +511,7 @@ class MasterPublicKeyMixin(ABC):
 class Xpub(MasterPublicKeyMixin):
 
     def __init__(self, *, derivation_prefix: str = None, root_fingerprint: str = None):
+        MasterPublicKeyMixin.__init__(self)
         self.xpub = None
         self.xpub_receive = None
         self.xpub_change = None
@@ -608,8 +619,7 @@ class Xpub(MasterPublicKeyMixin):
             self._derivation_prefix = derivation_prefix
         self.is_requesting_to_be_rewritten_to_wallet_file = True
 
-    @lru_cache(maxsize=None)
-    def derive_pubkey(self, for_change: int, n: int) -> bytes:
+    def _derive_pubkey(self, for_change: int, n: int) -> bytes:
         for_change = int(for_change)
         if for_change not in (0, 1):
             raise CannotDerivePubkey("forbidden path")
@@ -624,7 +634,7 @@ class Xpub(MasterPublicKeyMixin):
         return self.get_pubkey_from_xpub(xpub, (n,))
 
     @classmethod
-    def get_pubkey_from_xpub(self, xpub: str, sequence) -> bytes:
+    def get_pubkey_from_xpub(cls, xpub: str, sequence) -> bytes:
         node = BIP32Node.from_xkey(xpub).subkey_at_public_derivation(sequence)
         return node.eckey.get_public_key_bytes(compressed=True)
 
@@ -729,6 +739,7 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
     type = 'old'
 
     def __init__(self, d: dict):
+        MasterPublicKeyMixin.__init__(self)
         Deterministic_KeyStore.__init__(self, d)
         self.mpk = d.get('mpk')  # type: Optional[str]
         self._root_fingerprint = None
@@ -757,6 +768,10 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
         self.mpk = mpk
 
     def format_seed(self, seed):
+        """Returns seed in hex format.
+
+        seed: either in hex or as mnemonic words
+        """
         from . import old_mnemonic, mnemonic
         seed = mnemonic.normalize_text(seed)
         # see if seed was entered as hex
@@ -803,8 +818,7 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
         public_key = master_public_key + z*ecc.GENERATOR
         return public_key.get_public_key_bytes(compressed=False)
 
-    @lru_cache(maxsize=None)
-    def derive_pubkey(self, for_change, n) -> bytes:
+    def _derive_pubkey(self, for_change, n) -> bytes:
         for_change = int(for_change)
         if for_change not in (0, 1):
             raise CannotDerivePubkey("forbidden path")
@@ -1134,7 +1148,7 @@ def is_address_list(text: str) -> bool:
     return bool(parts) and all(bitcoin.is_address(x) for x in parts)
 
 
-def get_private_keys(text: str, *, allow_spaces_inside_key=True, raise_on_error=False) -> Optional[Sequence[str]]:
+def get_private_keys(text: str, *, allow_spaces_inside_key=True, raise_on_error=False) -> Sequence[str]:
     if allow_spaces_inside_key:  # see #1612
         parts = text.split('\n')
         parts = map(lambda x: ''.join(x.split()), parts)
@@ -1143,7 +1157,7 @@ def get_private_keys(text: str, *, allow_spaces_inside_key=True, raise_on_error=
         parts = text.split()
     if bool(parts) and all(bitcoin.is_private_key(x, raise_on_error=raise_on_error) for x in parts):
         return parts
-    return None
+    return []
 
 
 def is_private_key_list(text: str, *, allow_spaces_inside_key: bool = True, raise_on_error: bool = False) -> bool:
