@@ -42,6 +42,7 @@ from electrum.util import log_exceptions, ca_path, OldTaskGroup, get_asyncio_loo
     get_running_loop
 from electrum.invoices import Invoice, Request, PR_UNKNOWN, PR_PAID, BaseInvoice, PR_INFLIGHT
 from electrum import constants
+from electrum.lnutil import RECEIVED
 
 if TYPE_CHECKING:
     from aiohttp_socks import ProxyConnector
@@ -335,7 +336,6 @@ class NWCServer(Logger, EventListener):
                 content = json.loads(content)
                 if not isinstance(content, dict):
                     raise Exception("malformed content, not dict")
-                event.content = content
                 params: dict = content['params']
                 if not isinstance(params, dict):
                     raise Exception("malformed params, not dict")
@@ -362,30 +362,36 @@ class NWCServer(Logger, EventListener):
             elif method == "list_transactions":
                 task = self.handle_list_transactions(event, params)
             else:
-                self.logger.debug(f"Unsupported nwc method requested: {content.get('method')}")
-                await self.send_error(event, "NOT_IMPLEMENTED", f"{method} not supported")
+                self.logger.debug(f"Unsupported nwc method requested: {method}")
+                await self.send_error(event, "NOT_IMPLEMENTED", f"{method} not supported", error_restype=method)
                 continue
 
             if task:
-                await self.taskgroup.spawn(self.run_request_task(task, event))
+                await self.taskgroup.spawn(self.run_request_task(task, request_event=event, request_method=method))
 
-    async def run_request_task(self, task: Awaitable, request_event: nEvent) -> None:
+    async def run_request_task(self, task: Awaitable, *, request_event: nEvent, request_method: str = None) -> None:
         """Catches request handling exceptions and send an error response"""
         try:
             await task
         except Exception as e:
             self.logger.exception("Error handling nwc request")
-            await self.send_error(request_event, "INTERNAL", f"Error handling request: {str(e)[:100]}")
+            await self.send_error(
+                request_event, "INTERNAL", f"Error handling request: {str(e)[:100]}",
+                error_restype=request_method,
+            )
 
-    async def send_error(self, causing_event: nEvent, error_type: str, error_msg: str = "") -> None:
+    async def send_error(
+        self,
+        causing_event: nEvent,
+        error_type: str,
+        error_msg: str = "",
+        *,
+        error_restype: str = None,
+    ) -> None:
         """Sends an error as response to the passed nEvent, containing the error type and message"""
         to_pubkey_hex = causing_event.pubkey
         response_to_id = causing_event.id
-        res_type = None
-        if isinstance(causing_event.content, dict):  # we have replaced the content with the decrypted content
-            if 'method' in causing_event.content:
-                res_type = causing_event.content['method']
-        content = self.get_error_response(error_type, error_msg, res_type)
+        content = self.get_error_response(error_type, error_msg, error_restype)
         await self.send_encrypted_response(to_pubkey_hex, json.dumps(content), response_to_id)
 
     @staticmethod
@@ -480,12 +486,11 @@ class NWCServer(Logger, EventListener):
             address=None
         )
         req: Request = self.wallet.get_request(key)
+        info = self.wallet.lnworker.get_payment_info(req.payment_hash, direction=RECEIVED)
         try:
             lnaddr, b11 = self.wallet.lnworker.get_bolt11_invoice(
-                payment_hash=req.payment_hash,
-                amount_msat=amount_msat,
+                payment_info=info,
                 message=description,
-                expiry=expiry,
                 fallback_address=None
             )
         except Exception:
@@ -538,11 +543,10 @@ class NWCServer(Logger, EventListener):
             b11 = invoice.lightning_invoice
         elif self.wallet.get_request(invoice.rhash):
             direction = "incoming"
+            info = self.wallet.lnworker.get_payment_info(invoice.payment_hash, direction=RECEIVED)
             _, b11 = self.wallet.lnworker.get_bolt11_invoice(
-                payment_hash=bytes.fromhex(invoice.rhash),
-                amount_msat=invoice.amount_msat,
+                payment_info=info,
                 message=invoice.message,
-                expiry=invoice.exp,
                 fallback_address=None
             )
 
@@ -749,11 +753,10 @@ class NWCServer(Logger, EventListener):
         request: Optional[Request] = self.wallet.get_request(key)
         if not request or not request.is_lightning() or not status == PR_PAID:
             return
+        info = self.wallet.lnworker.get_payment_info(request.payment_hash, direction=RECEIVED)
         _, b11 = self.wallet.lnworker.get_bolt11_invoice(
-            payment_hash=request.payment_hash,
-            amount_msat=request.get_amount_msat(),
+            payment_info=info,
             message=request.message,
-            expiry=request.exp,
             fallback_address=None
         )
 
@@ -950,7 +953,8 @@ class NWCServer(Logger, EventListener):
         payments = self.wallet.lnworker.get_payments(status='settled')
         plist = payments.get(payment_hash)
         if plist:
-            info = self.wallet.lnworker.get_payment_info(payment_hash)
+            direction = plist[0].direction
+            info = self.wallet.lnworker.get_payment_info(payment_hash, direction=direction)
             if info:
                 dir, amount, fee, ts = self.wallet.lnworker.get_payment_value(info, plist)
                 fee = abs(fee) if fee else None
